package nts.parser;

import java.lang.System;
import java.util.*;

import org.antlr.runtime.Token;

import nts.interf.base.*;

public class VarTable implements IVarTable {
	
	public void accept(IVisitor v) { v.visit(this); }
	
	private Map<String, IVarTableEntry> table = new HashMap<String, IVarTableEntry>();
	private VarTable parent;
	private IScope scope;
	
	void parent(VarTable aParent) { parent = aParent; }
	
	// interface
	public IVarTable parent() { return parent; }
	public IScope scope() { return scope; }
	public List<IVarTableEntry> innermost() { 
		return new LinkedList<IVarTableEntry>(table.values()); 
	}
	public List<IVarTableEntry> innermostUnprimed() {
		List<IVarTableEntry> ret = new LinkedList<IVarTableEntry>();
		for (IVarTableEntry e : table.values()) {
			if (!e.isPrimed())
				ret.add(e);
		}
		return ret;
	}
	public List<IVarTableEntry> visible() {
		List<IVarTableEntry> ret = new LinkedList<IVarTableEntry>();
		VarTable ptr = this;
		while (ptr != null) {
			ret.addAll(ptr.innermost());
			ptr = ptr.parent;
		}
		return ret;
	}
	public List<IVarTableEntry> visibleUnprimed() {
		List<IVarTableEntry> ret = new LinkedList<IVarTableEntry>();
		VarTable ptr = this;
		while (ptr != null) {
			ret.addAll(ptr.innermostUnprimed());
			ptr = ptr.parent;
		}
		return ret;
	}
	
	public VarTable(VarTable aParent) {
		parent = aParent;
	}
	void setScope(IScope aScope) { scope = aScope; }
	
	public String toString() {
		String aux = table.keySet().toString();
		if (parent != null) {
			aux += "-->"+parent.toString();
		}
		return aux;
	}
	
	// return the entry for the given key if it exists or null
	public VarTableEntry get(String key) {
		VarTable ptr = this;
		while (ptr != null) {
			VarTableEntry ret = (VarTableEntry)ptr.table.get(key);
			if (ret != null)
				return ret;
			ptr = ptr.parent;
		}
		return null;
	}
	public boolean contains(String key) {
		return get(key) != null;
	}
	
	void declare(List<VarTableEntry> entries) {
		for (VarTableEntry e : entries) {
			declare(e);
		}
	}
	
	// adds a new entry, check for re-declaration
	// (variable hiding from different scopes not allowed)
	void declare(VarTableEntry entry) {
		
		entry.setVarTable(this);
		
		add(entry.name(), entry);
		
		// add entry for the primed version
		// (not for logical variables, parameters and TID)
		if (!entry.isLogical()) {
			switch (entry.modifier()) {
			case NO:
			case IN:
			case OUT:
			//case PARAM:
			//case LOGICAL:
			//case TID:
				VarTableEntry e2 = entry.createPrimed();
				add(e2.name(), e2);
			}
		}
	}
	private static void errRedeclaration(Token t1, Token t2) {
		System.err.println("Semantic error at "+Common.at(t1)+": Redeclaration of variable '"+t1.getText()+"' (already declared at "+Common.at(t2)+").");
		System.exit(1);
	}
	// adds and copies the scope
	private void add(String key, VarTableEntry entry) {
		IVarTableEntry e;
		if ((e = get(key)) != null) {
			Token t = entry.nameToken();
			Token t2 = ((VarTableEntry)e).nameToken();
			errRedeclaration(t,t2);
		}
		table.put(key, entry);
		entry.setScope(scope);
	}
	
	VarTableEntry checkDeclared(Token aT) {
		String s = aT.getText();
		VarTableEntry ret = get(s);
		
		// check if declared
		if (ret == null) {
			
			// to give better error messages ( = avoid "not declared" errors)
			// check that parameters, tid, logical variables are not accessed via primed versions
			if (Common.unprimed(s)) {
				VarTableEntry aux = get(Common.unprime(s));
				if (aux != null) {
					System.err.println("Semantic error at "+Common.at(aT)+": Primed access (\""+aT.getText()+"\") for parameters, tid, and logical vaiables is illegal.");
					System.exit(1);
				}
			}
			
			System.err.println("Semantic error at "+Common.at(aT)+": Variable \""+Common.unprime(aT.getText())+"\" is not declared.");
			System.exit(1);
		}
		
		return ret;
	}
	
	private static SemFlags sfSize = SemFlags.createArraySizeSpec();
	private void checkArrSize() {
		for (IVarTableEntry ee : this.innermost()) {
			VarTableEntry e = (VarTableEntry)ee;
			if (e.type().isArray()) {
				if (	(e.modifier() == EModifier.IN || e.modifier() == EModifier.OUT) && 
						!e.type().isRefToArray() ) {
					System.err.println("Semantic error at "+Common.at(e.nameToken())+": Input and output array parameters must be pure array references.");
					System.exit(1);
				}
				//SemFlags s = SemFlags.create(SemFlags.fOnlyParam | SemFlags.fNoPrime | SemFlags.fIsInIndexTerm);
				for (IExpr eex : e.size()) {
					Expr ex = (Expr)eex;
					ex.semanticChecks(this, sfSize);
//					ex.checkArrDefSize(this);
				}
			}
		}
	}
	
	// check if variables in this scope are not hiding variables in super-scopes
	// (important for logical variables in global NTS / subsystem annotations -- which are specified before declaration of global / local variables)
	void checkRedeclaration(VarTable aParent) {
		
		if (aParent == parent)
			return;
		
		parent = aParent;
		
		VarTableEntry e2;
		for (IVarTableEntry ei : table.values()) {
			VarTableEntry e = (VarTableEntry) ei;
			if (e.isPrimed())
				continue;
			String s = e.name();
			if ((e2 = parent.get(s)) != null) {
				errRedeclaration(e.nameToken(),e2.nameToken());
			}
		}
	}
	
	//private static SemFlags sfGen = SemFlags.createGeneral();
	public void semanticChecks() {
		// annotations
		for (IVarTableEntry e : table.values()) {
			((VarTableEntry)e).semanticChecks(this);
		}
		
		checkArrSize();
	}
}
